"""
Import an SBML string from a file and visualize it to a network on canvas.
Version 0.02: Author: Jin Xu (2021)
"""


# pylint: disable=maybe-no-member

from ast import Num
from inspect import Parameter
from ntpath import join
from re import S
import wx
from wx.core import CENTER
from rkviewer.canvas.data import TEXT_POSITION_CHOICES, TextAlignment, TextPosition
from rkviewer.plugin.classes import PluginMetadata, WindowedPlugin, PluginCategory
from rkviewer.plugin import api
from rkviewer.plugin.api import Node, Vec2, Reaction, Color
import os
import simplesbml # does not have to import in the main.py too
from libsbml import *
import math
import random as _random

class IMPORTSBML(WindowedPlugin):
    metadata = PluginMetadata(
        name='ImportSBML',
        author='Jin Xu',
        version='0.5.8',
        short_desc='Import SBML.',
        long_desc='Import an SBML String from a file and visualize it as a network on canvas.',
        category=PluginCategory.ANALYSIS
    )

    def create_window(self, dialog):
        """
        Create a window to import SBML.
        Args:
            self
            dialog
        """
        self.window = wx.Panel(dialog, pos=(5,100), size=(300, 320))
        self.sbmlStr = ''
        show_btn = wx.Button(self.window, -1, 'Load', (5, 5))
        show_btn.Bind(wx.EVT_BUTTON, self.Show)

        copy_btn = wx.Button(self.window, -1, 'Copy To Clipboard', (83, 5))
        copy_btn.Bind(wx.EVT_BUTTON, self.Copy)

        visualize_btn = wx.Button(self.window, -1, 'Visualize', (205, 5))
        visualize_btn.Bind(wx.EVT_BUTTON, self.Visualize)

        wx.StaticText(self.window, -1, 'SBML string:', (5,30))
        self.SBMLText = wx.TextCtrl(self.window, -1, "", (10, 50), size=(260, 220), style=wx.TE_MULTILINE|wx.HSCROLL)
        self.SBMLText.SetInsertionPoint(0)

        return self.window

    def Show(self, evt):
        """
        Handler for the "Import" button.
        Open the SBML file and show it in the TextCtrl box.
        """

        self.dirname=""  #set directory name to blank
        dlg = wx.FileDialog(self.window, "Choose a file to open", self.dirname, wildcard="SBML files (*.xml)|*.xml", style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST) #open the dialog boxto open file
        if dlg.ShowModal() == wx.ID_OK:  #if positive button selected....
            self.filename = dlg.GetFilename()  #get the filename of the file
            self.dirname = dlg.GetDirectory()  #get the directory of where file is located
            f = open(os.path.join(self.dirname, self.filename), 'r')  #traverse the file directory and find filename in the OS
            self.sbmlStr = f.read()
            self.SBMLText.SetValue(f.read())  #open the file from location as read
            self.SBMLText.WriteText(self.sbmlStr)
            f.close
        dlg.Destroy()

    def Copy(self, evt):
        """
        Handler for the "Copy" button.
        Copy the SBML string to a clipboard.
        """
        self.dataObj = wx.TextDataObject()
        self.dataObj.SetText(self.SBMLText.GetValue())
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(self.dataObj)
            wx.TheClipboard.Close()
        else:
            wx.MessageBox("Unable to open the clipboard", "Error")

    def Visualize(self, evt):
        """
        Handler for the "Visualize" button.
        Visualize the SBML string to a network shown on the canvas.
        """  

        with wx.BusyCursor():
        #with wx.BusyInfo("Please wait, working..."):
            self.DisplayModel(self.sbmlStr, True, False)


    def DisplayModel(self, sbmlStr, showDialogues, useSeed):
        """
        Visualize an SBML string as a network shown on the canvas.
        Args:
          self
          document: SBMLDocument object created from the sbml string
          sbmlStr: sbml string to display
          showDialogues: if false, hides pop-up windows
          useSeed: if true, constant seed for random number generation used,
                   ensuring that different visualizations created from the same
                   file will always have the same layout
        """
        if useSeed:
          _random.seed(13)

        def hex_to_rgb(value):
            value = value.lstrip('#')
            if len(value) == 6:
                value = value + 'ff'
            return tuple(int(value[i:i+2], 16) for i in (0, 2, 4, 6))

        # if len(sbmlStr) == 0:
        #   if showDialogues:
        #     wx.MessageBox("Please import an SBML file.", "Message", wx.OK | wx.ICON_INFORMATION)
        # else:
        if len(sbmlStr) != 0:
            net_index = 0
            api.clear_network(net_index)
            comp_id_list = []
            comp_dimension_list = []
            comp_position_list = []
            spec_id_list = []
            specGlyph_id_list = []
            spec_specGlyph_id_list = []
            spec_dimension_list = []
            spec_position_list = []
            spec_text_alignment_list = []
            spec_text_position_list = []
            spec_concentration_list = []

            comp_render = []
            spec_render = []
            rxn_render = []

            shapeIdx = 0

            
            #set the default values without render info:
            comp_fill_color = (158, 169, 255, 200)
            comp_border_color = (0, 29, 255, 255)
            comp_border_width = 2.0
            spec_fill_color = (255, 204, 153, 200)
            spec_border_color = (255, 108, 9, 255)
            spec_border_width = 2.0
            reaction_line_color = (91, 176, 253, 255)
            reaction_line_width = 3.0
            #text_line_color = (0, 0, 0, 255)
            #text_line_width = 1.

            try: #possible invalid sbml
                ### from here for layout ###
                document = readSBMLFromString(sbmlStr)
                model_layout = document.getModel()
                mplugin = (model_layout.getPlugin("layout"))
                
                # Get the first Layout object via LayoutModelPlugin object.
                #
                # if mplugin is None:
                #     if showDialogues:
                #         wx.MessageBox("There is no layout information, so positions are randomly assigned.", "Message", wx.OK | wx.ICON_INFORMATION)
                # else:
                if mplugin is not None:
                    layout = mplugin.getLayout(0)
                    # if layout is None:
                    #     if showDialogues:
                    #         wx.MessageBox("There is no layout information, so positions are randomly assigned.", "Message", wx.OK | wx.ICON_INFORMATION)
                    # else:
                    if layout is not None:
                        numCompGlyphs = layout.getNumCompartmentGlyphs()
                        numSpecGlyphs = layout.getNumSpeciesGlyphs()
                        numReactionGlyphs = layout.getNumReactionGlyphs()

                        for i in range(numCompGlyphs):
                            compGlyph = layout.getCompartmentGlyph(i)
                            temp_id = compGlyph.getCompartmentId()
                            comp_id_list.append(temp_id)
                            boundingbox = compGlyph.getBoundingBox()
                            height = boundingbox.getHeight()
                            width = boundingbox.getWidth()
                            pos_x = boundingbox.getX()
                            pos_y = boundingbox.getY()
                            comp_dimension_list.append([width,height])
                            comp_position_list.append([pos_x,pos_y])


                        reaction_id_list = []
                        reaction_center_list = []
                        kinetics_list = []
                        #rct_specGlyph_list = []
                        #prd_specGlyph_list = []
                        reaction_center_handle_list = []
                        rct_specGlyph_handle_list = []
                        prd_specGlyph_handle_list = []
                        reaction_mod_list = []
                        mod_specGlyph_list = []

                        for i in range(numReactionGlyphs):
                            reactionGlyph = layout.getReactionGlyph(i)
                            curve = reactionGlyph.getCurve()
                            # listOfCurveSegments = curve.getListOfCurveSegments()
                            # for j in range(len(listOfCurveSegments)):
                            #     #center_x = curve.getCurveSegment(j).getStart().x()
                            #     #center_y = curve.getCurveSegment(j).getStart().y()
                            #     center_x = curve.getCurveSegment(j).getStart().getXOffset()
                            #     center_y = curve.getCurveSegment(j).getStart().getYOffset()
                            for segment in curve.getListOfCurveSegments():
                                center_x = segment.getStart().getXOffset()
                                center_y = segment.getStart().getYOffset()
                                reaction_center_list.append([center_x, center_y])
                            reaction_id = reactionGlyph.getReactionId()
                            reaction_id_list.append(reaction_id)
                            reaction = model_layout.getReaction(reaction_id)
                            kinetics = reaction.getKineticLaw().getFormula()
                            kinetics_list.append(kinetics)

                            temp_mod_list = []
                            for j in range(len(reaction.getListOfModifiers())):
                                modSpecRef = reaction.getModifier(j)
                                temp_mod_list.append(modSpecRef.getSpecies())
                            reaction_mod_list.append(temp_mod_list)

                            numSpecRefGlyphs = reactionGlyph.getNumSpeciesReferenceGlyphs()

                            #rct_specGlyph_temp_list = []
                            #prd_specGlyph_temp_list = []
                            rct_specGlyph_handles_temp_list = []
                            prd_specGlyph_handles_temp_list = []  
                            mod_specGlyph_temp_list = []

                            for j in range(numSpecRefGlyphs):
                                alignment_name = TextAlignment.CENTER
                                position_name = TextPosition.IN_NODE
                                specRefGlyph = reactionGlyph.getSpeciesReferenceGlyph(j)
                                #specRefGlyph_id = specRefGlyph.get
                                # 
                                # Id()

                                curve = specRefGlyph.getCurve()                             
                                for segment in curve.getListOfCurveSegments():
                                        # print(segment.getStart().getXOffset())
                                        # print(segment.getStart().getYOffset())
                                        # print(segment.getEnd().getXOffset())
                                        # print(segment.getEnd().getYOffset())
                                        try:
                                            center_handle = [segment.getBasePoint1().getXOffset(), 
                                                        segment.getBasePoint1().getYOffset()]                                
                                            spec_handle = [segment.getBasePoint2().getXOffset(),
                                                    segment.getBasePoint2().getYOffset()]
                                        except:
                                            center_handle = []
                                            spec_handle = []

                                role = specRefGlyph.getRoleString()
                                specGlyph_id = specRefGlyph.getSpeciesGlyphId()
                                specGlyph = layout.getSpeciesGlyph(specGlyph_id)
                                
                                for k in range(numSpecGlyphs):
                                    textGlyph_temp = layout.getTextGlyph(k)
                                    temp_specGlyph_id = textGlyph_temp.getOriginOfTextId()
                                    if temp_specGlyph_id == specGlyph_id:
                                        textGlyph = textGlyph_temp

                                spec_id = specGlyph.getSpeciesId()
                                spec = model_layout.getSpecies(spec_id)
                                
                                try:
                                    concentration = spec.getInitialConcentration()
                                except:
                                    concentration = 1.
                                spec_boundingbox = specGlyph.getBoundingBox()
                                height = spec_boundingbox.getHeight()
                                width = spec_boundingbox.getWidth()
                                pos_x = spec_boundingbox.getX()
                                pos_y = spec_boundingbox.getY()

                                try:
                                    text_boundingbox = textGlyph.getBoundingBox()
                                    text_pos_x = text_boundingbox.getX()
                                    text_pos_y = text_boundingbox.getY()
                                    if text_pos_x < pos_x:
                                        alignment_name = TextAlignment.LEFT
                                    if text_pos_x > pos_x:
                                        alignment_name = TextAlignment.RIGHT  
                                    if text_pos_y < pos_y:
                                        position_name = TextPosition.ABOVE
                                    if text_pos_y > pos_y:
                                        position_name = TextPosition.BELOW
                                    if text_pos_y == pos_y and text_pos_x != pos_x:
                                        position_name = TextPosition.NEXT_TO 
                                except:
                                    pass   

                                if specGlyph_id not in specGlyph_id_list:
                                    spec_id_list.append(spec_id)
                                    specGlyph_id_list.append(specGlyph_id)
                                    spec_specGlyph_id_list.append([spec_id,specGlyph_id])
                                    spec_dimension_list.append([width,height])
                                    spec_position_list.append([pos_x,pos_y])
                                    spec_text_alignment_list.append(alignment_name)
                                    spec_text_position_list.append(position_name)
                                    spec_concentration_list.append(concentration)
                                

                                if role == "substrate": #it is a rct
                                    #rct_specGlyph_temp_list.append(specGlyph_id)
                                    rct_specGlyph_handles_temp_list.append([specGlyph_id,spec_handle])
                                elif role == "product": #it is a prd
                                    #prd_specGlyph_temp_list.append(specGlyph_id)
                                    prd_specGlyph_handles_temp_list.append([specGlyph_id,spec_handle])
                                elif role == "modifier": #it is a modifier
                                    mod_specGlyph_temp_list.append(specGlyph_id)
                            #rct_specGlyph_list.append(rct_specGlyph_temp_list)
                            #prd_specGlyph_list.append(prd_specGlyph_temp_list)
                            reaction_center_handle_list.append(center_handle)
                            rct_specGlyph_handle_list.append(rct_specGlyph_handles_temp_list)
                            prd_specGlyph_handle_list.append(prd_specGlyph_handles_temp_list)    
                            mod_specGlyph_list.append(mod_specGlyph_temp_list)

                        #orphan nodes
                        for i in range(numSpecGlyphs):
                            specGlyph = layout.getSpeciesGlyph(i)
                            specGlyph_id = specGlyph.getId()
                            if specGlyph_id not in specGlyph_id_list:
                                specGlyph_id_list.append(specGlyph_id)
                                spec_id = specGlyph.getSpeciesId()
                                spec_id_list.append(spec_id)
                                spec_specGlyph_id_list.append([spec_id,specGlyph_id])
                                boundingbox = specGlyph.getBoundingBox()
                                height = boundingbox.getHeight()
                                width = boundingbox.getWidth()
                                pos_x = boundingbox.getX()
                                pos_y = boundingbox.getY()
                                spec_dimension_list.append([width,height])
                                spec_position_list.append([pos_x,pos_y])
                                try:
                                    concentration = spec.getInitialConcentration()
                                except:
                                    concentration = 1.
                                spec_concentration_list.append(concentration)
                                alignment_name = TextAlignment.CENTER
                                position_name = TextPosition.IN_NODE
                                for k in range(numSpecGlyphs):
                                    textGlyph_temp = layout.getTextGlyph(k)
                                    temp_specGlyph_id = textGlyph_temp.getOriginOfTextId()
                                    if temp_specGlyph_id == specGlyph_id:
                                        textGlyph = textGlyph_temp
                                try:
                                    text_boundingbox = textGlyph.getBoundingBox()
                                    text_pos_x = text_boundingbox.getX()
                                    text_pos_y = text_boundingbox.getY()
                                    if text_pos_x < pos_x:
                                        alignment_name = TextAlignment.LEFT
                                    if text_pos_x > pos_x:
                                        alignment_name = TextAlignment.RIGHT  
                                    if text_pos_y < pos_y:
                                        position_name = TextPosition.ABOVE
                                    if text_pos_y > pos_y:
                                        position_name = TextPosition.BELOW
                                    if text_pos_y == pos_y and text_pos_x != pos_x:
                                        position_name = TextPosition.NEXT_TO 
                                except:
                                    pass
                                spec_text_alignment_list.append(alignment_name)
                                spec_text_position_list.append(position_name)

                        
                        #print(reaction_mod_list)
                        #print(mod_specGlyph_list)
                        #print(spec_specGlyph_id_list)
                        rPlugin = layout.getPlugin("render")
                        if (rPlugin != None and rPlugin.getNumLocalRenderInformationObjects() > 0):
                        #if rPlugin != None:
                            #wx.MessageBox("The diversity of each graphical object is not shown.", "Message", wx.OK | wx.ICON_INFORMATION)
                            info = rPlugin.getRenderInformation(0)
                            color_list = []
                            # comp_render = []
                            # spec_render = []
                            # rxn_render = []
                            # text_render = []
                            # for j in range(0, info.getNumLineEndings()):
                            #     LineEnding = info.getLineEndings(j)

                            for  j in range (0, info.getNumColorDefinitions()):
                                color = info.getColorDefinition(j)
                                color_list.append([color.getId(),color.createValueString()])

                            for j in range (0, info.getNumStyles()):
                                style = info.getStyle(j)
                                group = style.getGroup()
                                # for element in group.getListOfElements(): 
                                #     print(element.getElementName(), element.getStroke())
                                typeList = style.createTypeString()
                                idList = style.createIdString()
                                if 'COMPARTMENTGLYPH' in typeList:
                                    for k in range(len(color_list)):
                                        if color_list[k][0] == group.getFill():
                                            comp_fill_color = hex_to_rgb(color_list[k][1])
                                        if color_list[k][0] == group.getStroke():
                                            comp_border_color = hex_to_rgb(color_list[k][1])
                                    comp_border_width = group.getStrokeWidth()
                                    comp_render.append([idList,comp_fill_color,comp_border_color,comp_border_width])
                                elif 'SPECIESGLYPH' in typeList:
                                    for k in range(len(color_list)):
                                        if color_list[k][0] == group.getFill():
                                            spec_fill_color = hex_to_rgb(color_list[k][1])
                                        if color_list[k][0] == group.getStroke():
                                            spec_border_color = hex_to_rgb(color_list[k][1])
                                    spec_border_width = group.getStrokeWidth()
                                    name_list = []
                                    for element in group.getListOfElements():
                                        name = element.getElementName()
                                        name_list.append(name)
                                        try:
                                            NumRenderpoints = element.getListOfElements().getNumRenderPoints()
                                        except:
                                            NumRenderpoints = 0
                                    if name == "ellipse": #circel and text-outside
                                        shapeIdx = 1
                                    elif name == "polygon" and NumRenderpoints == 6:
                                        shapeIdx = 2
                                    elif name == "polygon" and NumRenderpoints == 2:
                                        shapeIdx = 3
                                    elif name == "polygon" and NumRenderpoints == 3:
                                        shapeIdx = 4
                                    elif name == "rectangle" and spec_fill_color == '#ffffffff' and spec_border_color == '#ffffffff':
                                        shapeIdx = 5
                                    #elif name == "ellipse" and flag_text_out == 1:
                                    #    shapeIdx = 6
                                    else: # name == "rectangle"/demo combo/others as default (rectangle)
                                        shapeIdx = 0
                                    spec_render.append([idList,spec_fill_color,spec_border_color,spec_border_width,shapeIdx])

                                elif 'REACTIONGLYPH' in typeList:
                                    for k in range(len(color_list)):
                                        if color_list[k][0] == group.getStroke():
                                            reaction_line_color = hex_to_rgb(color_list[k][1])
                                    reaction_line_width = group.getStrokeWidth()
                                    rxn_render.append([idList, reaction_line_color,reaction_line_width])

                model = simplesbml.loadSBMLStr(sbmlStr)

                numFloatingNodes  = model.getNumFloatingSpecies()
                FloatingNodes_ids = model.getListOfFloatingSpecies()
                numBoundaryNodes  = model.getNumBoundarySpecies()
                BoundaryNodes_ids = model.getListOfBoundarySpecies()
                numRxns   = model.getNumReactions()
                Rxns_ids  = model.getListOfReactionIds()
                numComps  = model.getNumCompartments()
                Comps_ids = model.getListOfCompartmentIds()
                numNodes = numFloatingNodes + numBoundaryNodes

                parameter_list = model.getListOfParameterIds()

                for p in parameter_list:
                    if model.isParameterValueSet(p): 
                    #if there is only parameter id without parameter value, it won't be considered 
                        api.set_parameter_value(net_index, p, model.getParameterValue(p))

                comp_node_list = [0]*numComps #Note: numComps is different from numCompGlyphs
                for i in range(numComps):
                    comp_node_list[i] = []

            
                #if there is layout info:
                if len(spec_id_list) != 0:
                    for i in range(numComps):
                        temp_id = Comps_ids[i]
                        vol= model.getCompartmentVolume(i)
                        if temp_id == "_compartment_default_":
                            api.add_compartment(net_index, id=temp_id, volume = vol,
                            size=Vec2(3900,2400), position=Vec2(10,10),
                            fill_color = api.Color(255, 255, 255, 255), #the last digit for transparent
                            border_color = api.Color(255, 255, 255, 255),
                            border_width = comp_border_width)
                        else:
                            if len(comp_id_list) != 0:
                            #if mplugin is not None:                    
                                for j in range(numCompGlyphs):
                                    if comp_id_list[j] == temp_id:
                                        dimension = comp_dimension_list[j]
                                        position = comp_position_list[j]
                                for j in range(len(comp_render)):
                                    if temp_id == comp_render[j][0]:
                                        comp_fill_color = comp_render[j][1]
                                        comp_border_color = comp_render[j][2]
                                        comp_border_width = comp_render[j][3]

                            else:# no layout info about compartment,
                                # then the whole size of the canvas is the compartment size
                                # modify the compartment size using the max_rec function above
                                # random assigned network:
                                # dimension = [800,800]
                                # position = [40,40]
                                # the whole size of the compartment: 4000*2500
                                dimension = [3900,2400]
                                position = [10,10]
                                comp_fill_color = (255, 255, 255, 255) #the last digit for transparent
                                comp_border_color = (255, 255, 255, 255)
                    
                            api.add_compartment(net_index, id=temp_id, volume = vol,
                            size=Vec2(dimension[0],dimension[1]),position=Vec2(position[0],position[1]),
                            fill_color = api.Color(comp_fill_color[0],comp_fill_color[1],comp_fill_color[2],comp_fill_color[3]),
                            border_color = api.Color(comp_border_color[0],comp_border_color[1],comp_border_color[2],comp_border_color[3]),
                            border_width = comp_border_width)

                    id_list = []
                    nodeIdx_list = [] #get_nodes idx do not follow the same order of add_node
                    nodeIdx_specGlyph_list = []
                    nodeIdx_specGlyph_alias_list = []
                    numSpec_in_reaction = len(spec_specGlyph_id_list) 
                    # orphan nodes have been considered, so numSpec_in_reaction should equals to numSpecGlyphs
                    for i in range (numSpec_in_reaction):
                        temp_id = spec_specGlyph_id_list[i][0]
                        temp_concentration = spec_concentration_list[i]
                        tempGlyph_id = spec_specGlyph_id_list[i][1]
                        dimension = spec_dimension_list[i]
                        position = spec_position_list[i]
                        text_alignment = spec_text_alignment_list[i]
                        text_position = spec_text_position_list[i]
                        comp_id = model.getCompartmentIdSpeciesIsIn(temp_id)
                        for j in range(numFloatingNodes):
                            if temp_id == FloatingNodes_ids[j]:
                                if temp_id not in id_list:
                                    for k in range(len(spec_render)):
                                        if temp_id == spec_render[k][0]:
                                            spec_fill_color = spec_render[k][1]
                                            spec_border_color = spec_render[k][2]
                                            spec_border_width = spec_render[k][3]
                                            shapeIdx = spec_render[k][4]
                                    nodeIdx_temp = api.add_node(net_index, id=temp_id, floating_node = True,
                                    size=Vec2(dimension[0],dimension[1]), position=Vec2(position[0],position[1]),
                                    fill_color=api.Color(spec_fill_color[0],spec_fill_color[1],spec_fill_color[2],spec_fill_color[3]),
                                    border_color=api.Color(spec_border_color[0],spec_border_color[1],spec_border_color[2],spec_border_color[3]),
                                    border_width=spec_border_width, shape_index=shapeIdx, concentration = temp_concentration)
                                    api.set_node_shape_property(net_index, nodeIdx_temp, -1, "alignment", text_alignment)
                                    api.set_node_shape_property(net_index, nodeIdx_temp, -1, "position", text_position)
                                    id_list.append(temp_id)
                                    nodeIdx_list.append(nodeIdx_temp)
                                    nodeIdx_specGlyph_list.append([nodeIdx_temp,tempGlyph_id])
                                else:
                                    index = id_list.index(temp_id)
                                    nodeIdx_temp = api.add_alias(net_index, original_index=index,
                                    size=Vec2(dimension[0],dimension[1]), position=Vec2(position[0],position[1]) )
                                    api.set_node_shape_property(net_index, nodeIdx_temp, -1, "alignment", text_alignment)
                                    api.set_node_shape_property(net_index, nodeIdx_temp, -1, "position", text_position)
                                    id_list.append(temp_id)
                                    nodeIdx_list.append(nodeIdx_temp)
                                    nodeIdx_specGlyph_alias_list.append([nodeIdx_temp,tempGlyph_id])
                                for k in range(numCompGlyphs):
                                    if len(comp_id_list) !=0 and comp_id == comp_id_list[k]:
                                        comp_node_list[k].append(nodeIdx_temp)
                        for j in range(numBoundaryNodes):
                            if temp_id == BoundaryNodes_ids[j]:
                                if temp_id not in id_list:
                                    for k in range(len(spec_render)):
                                        if temp_id == spec_render[k][0]:
                                            spec_fill_color = spec_render[k][1]
                                            spec_border_color = spec_render[k][2]
                                            spec_border_width = spec_render[k][3]
                                            shapeIdx = spec_render[k][4]
                                    nodeIdx_temp = api.add_node(net_index, id=temp_id, floating_node = False,
                                    size=Vec2(dimension[0],dimension[1]), position=Vec2(position[0],position[1]),
                                    fill_color=api.Color(spec_fill_color[0],spec_fill_color[1],spec_fill_color[2],spec_fill_color[3]),
                                    border_color=api.Color(spec_border_color[0],spec_border_color[1],spec_border_color[2],spec_border_color[3]),
                                    border_width=spec_border_width, shape_index=shapeIdx, concentration = temp_concentration)
                                    api.set_node_shape_property(net_index, nodeIdx_temp, -1, "alignment", text_alignment)
                                    api.set_node_shape_property(net_index, nodeIdx_temp, -1, "position", text_position)
                                    id_list.append(temp_id)
                                    nodeIdx_list.append(nodeIdx_temp)
                                    nodeIdx_specGlyph_list.append([nodeIdx_temp,tempGlyph_id])
                                else:
                                    index = id_list.index(temp_id)
                                    nodeIdx_temp = api.add_alias(net_index, original_index=index,
                                    size=Vec2(dimension[0],dimension[1]), position=Vec2(position[0],position[1]))
                                    api.set_node_shape_property(net_index, nodeIdx_temp, -1, "alignment", text_alignment)
                                    api.set_node_shape_property(net_index, nodeIdx_temp, -1, "position", text_position)
                                    id_list.append(temp_id)
                                    nodeIdx_list.append(nodeIdx_temp)
                                    nodeIdx_specGlyph_alias_list.append([nodeIdx_temp,tempGlyph_id])
                                for k in range(numCompGlyphs):
                                    if len(comp_id) != 0 and comp_id == comp_id_list[k]:
                                        comp_node_list[k].append(nodeIdx_temp)

                    if len(comp_id_list) != 0:
                        for i in range(numComps):
                            temp_id = Comps_ids[i]
                            if temp_id == '_compartment_default_': 
                                #numNodes is different from len(nodeIdx_list) because of alias node
                                node_list_default = [item for item in range(len(nodeIdx_list))]
                                for j in range(len(node_list_default)):
                                    try:
                                        api.set_compartment_of_node(net_index=net_index, node_index=node_list_default[j], comp_index=i) 
                                    except:
                                        pass # Orphan nodes are removed
                            for j in range(numCompGlyphs):
                                if comp_id_list[j] == temp_id:
                                    node_list_temp = comp_node_list[j]
                                else:
                                    node_list_temp = []
                                for k in range(len(node_list_temp)):
                                    api.set_compartment_of_node(net_index=net_index, node_index=node_list_temp[k], comp_index=i)
                    else:
                        for i in range(len(nodeIdx_list)):
                            api.set_compartment_of_node(net_index=net_index, node_index=nodeIdx_list[i], comp_index=0)


                    nodeIdx_specGlyph_whole_list = nodeIdx_specGlyph_list + nodeIdx_specGlyph_alias_list

                    dummy_node_id_index = 0
                    for i in range (numReactionGlyphs):
                        src = []
                        dst = []
                        mod = []
                        src_handle = []
                        dst_handle = []
                        temp_id = reaction_id_list[i]
                        kinetics = kinetics_list[i]
                        rct_num = len(rct_specGlyph_handle_list[i])
                        prd_num = len(prd_specGlyph_handle_list[i])
                        mod_num = max(len(mod_specGlyph_list[i]),len(reaction_mod_list[i]))

                        # for j in range(rct_num):
                        #     temp_specGlyph_id = rct_specGlyph_list[i][j]
                        #     for k in range(numSpec_in_reaction):
                        #         if temp_specGlyph_id == nodeIdx_specGlyph_whole_list[k][1]:
                        #             rct_idx = nodeIdx_specGlyph_whole_list[k][0]
                                    
                        #     src.append(rct_idx)

                        # for j in range(prd_num):
                        #     temp_specGlyph_id = prd_specGlyph_list[i][j]
                        #     for k in range(numSpec_in_reaction):
                        #         if temp_specGlyph_id == nodeIdx_specGlyph_whole_list[k][1]:
                        #             prd_idx = nodeIdx_specGlyph_whole_list[k][0]
                        #     dst.append(prd_idx)

                        for j in range(rct_num):
                            temp_specGlyph_id = rct_specGlyph_handle_list[i][j][0]
                            for k in range(numSpec_in_reaction):
                                if temp_specGlyph_id == nodeIdx_specGlyph_whole_list[k][1]:
                                    rct_idx = nodeIdx_specGlyph_whole_list[k][0]
                            src.append(rct_idx)
                            src_handle.append(rct_specGlyph_handle_list[i][j][1])

                        for j in range(prd_num):
                            temp_specGlyph_id = prd_specGlyph_handle_list[i][j][0]
                            for k in range(numSpec_in_reaction):
                                if temp_specGlyph_id == nodeIdx_specGlyph_whole_list[k][1]:
                                    prd_idx = nodeIdx_specGlyph_whole_list[k][0]
                            dst.append(prd_idx)
                            dst_handle.append(prd_specGlyph_handle_list[i][j][1])

                        for j in range(mod_num):
                            if len(mod_specGlyph_list[i]) != 0:
                                temp_specGlyph_id = mod_specGlyph_list[i][j]
                                for k in range(numSpec_in_reaction):
                                    if temp_specGlyph_id == nodeIdx_specGlyph_whole_list[k][1]:
                                        mod_idx = nodeIdx_specGlyph_whole_list[k][0]
                                mod.append(mod_idx)
                            else:
                                for k in range(len(spec_specGlyph_id_list)):
                                    if reaction_mod_list[i][j] == spec_specGlyph_id_list[k][0]:
                                        temp_specGlyph_id = spec_specGlyph_id_list[k][1]
                                for k in range(numSpec_in_reaction):
                                    if temp_specGlyph_id == nodeIdx_specGlyph_whole_list[k][1]:
                                        mod_idx = nodeIdx_specGlyph_whole_list[k][0]
                                mod.append(mod_idx)

                        mod = set(mod)

                        for j in range(len(rxn_render)):
                            if temp_id == rxn_render[j][0]:
                                reaction_line_color = rxn_render[j][1]
                                reaction_line_width = rxn_render[j][2]

                        try:
                            src_corr = []
                            [src_corr.append(x) for x in src if x not in src_corr]
                            dst_corr = []
                            [dst_corr.append(x) for x in dst if x not in dst_corr]

                            center_position = reaction_center_list[i]
                            center_handle = reaction_center_handle_list[i]
                            handles = [center_handle]
                            handles.extend(src_handle)
                            handles.extend(dst_handle)
                            # print("rcts:", src_corr)
                            # print("prds:", dst_corr)                      
                            idx = api.add_reaction(net_index, id=temp_id, reactants=src_corr, products=dst_corr,
                            fill_color=api.Color(reaction_line_color[0],reaction_line_color[1],reaction_line_color[2],reaction_line_color[3]),
                            line_thickness=reaction_line_width, modifiers = mod)
                            api.update_reaction(net_index, idx, ratelaw = kinetics)
                            handles_Vec2 = []  
                            #print(handles)        
                            for i in range(len(handles)):
                                handles_Vec2.append(Vec2(handles[i][0],handles[i][1]))
                            api.update_reaction(net_index, idx, 
                            center_pos = Vec2(center_position[0],center_position[1]), 
                            handle_positions=handles_Vec2, 
                            fill_color=api.Color(reaction_line_color[0],reaction_line_color[1],reaction_line_color[2],reaction_line_color[3]))

                        except: #There is no info about the center/handle positions, so set as default 
                            src_corr = []
                            [src_corr.append(x) for x in src if x not in src_corr]
                            dst_corr = []
                            [dst_corr.append(x) for x in dst if x not in dst_corr]
                            idx = api.add_reaction(net_index, id=temp_id, reactants=src_corr, products=dst_corr,
                            fill_color=api.Color(reaction_line_color[0],reaction_line_color[1],reaction_line_color[2],reaction_line_color[3]),
                            line_thickness=reaction_line_width, modifiers = mod)
                            api.update_reaction(net_index, idx, ratelaw = kinetics,
                            fill_color=api.Color(reaction_line_color[0],reaction_line_color[1],reaction_line_color[2],reaction_line_color[3]))
                
                else: # there is no layout information, assign position randomly and size as default
                    comp_id_list = Comps_ids

                    for i in range(numComps):
                        temp_id = Comps_ids[i]
                        vol= model.getCompartmentVolume(i)
                        dimension = [3900,2400]
                        position = [10,10]

                        api.add_compartment(net_index, id=temp_id, volume = vol,
                        size=Vec2(dimension[0],dimension[1]),position=Vec2(position[0],position[1]),
                        fill_color = api.Color(comp_fill_color[0],comp_fill_color[1],comp_fill_color[2],comp_fill_color[3]),
                        border_color = api.Color(comp_border_color[0],comp_border_color[1],comp_border_color[2],comp_border_color[3]),
                        border_width = comp_border_width)

                    for i in range (numFloatingNodes):
                        temp_id = FloatingNodes_ids[i]
                        comp_id = model.getCompartmentIdSpeciesIsIn(temp_id)
                        try:
                            temp_concentration = model.getSpeciesInitialConcentration(temp_id)
                        except:
                            temp_concentration = 1.0
                        nodeIdx_temp = api.add_node(net_index, id=temp_id, size=Vec2(60,40), floating_node = True,
                        position=Vec2(40 + math.trunc (_random.random()*800), 40 + math.trunc (_random.random()*800)),
                        fill_color=api.Color(spec_fill_color[0],spec_fill_color[1],spec_fill_color[2],spec_fill_color[3]),
                        border_color=api.Color(spec_border_color[0],spec_border_color[1],spec_border_color[2],spec_border_color[3]),
                        border_width=spec_border_width, shape_index=shapeIdx, concentration=temp_concentration)
                        for j in range(numComps):
                            if comp_id == comp_id_list[j]:
                                comp_node_list[j].append(nodeIdx_temp)

                    for i in range (numBoundaryNodes):
                        temp_id = BoundaryNodes_ids[i]
                        comp_id = model.getCompartmentIdSpeciesIsIn(temp_id)
                        try:
                            temp_concentration = model.getSpeciesInitialConcentration(temp_id)
                        except:
                            temp_concentration = 1.0
                        nodeIdx_temp = api.add_node(net_index, id=temp_id, size=Vec2(60,40), floating_node = False,
                        position=Vec2(40 + math.trunc (_random.random()*800), 40 + math.trunc (_random.random()*800)),
                        fill_color=api.Color(spec_fill_color[0],spec_fill_color[1],spec_fill_color[2],spec_fill_color[3]),
                        border_color=api.Color(spec_border_color[0],spec_border_color[1],spec_border_color[2],spec_border_color[3]),
                        border_width=spec_border_width, shape_index=shapeIdx, concentration=temp_concentration)
                        for j in range(numComps):
                            if comp_id == comp_id_list[j]:
                                comp_node_list[j].append(nodeIdx_temp)


                    for i in range(numComps):
                        temp_id = Comps_ids[i]
                        for j in range(numComps):
                            if comp_id_list[j] == temp_id:
                                node_list_temp = comp_node_list[j]
                            for k in range(len(node_list_temp)):
                                api.set_compartment_of_node(net_index=net_index, node_index=node_list_temp[k], comp_index=i)

                    numNodes = api.node_count(net_index)
                    allNodes = api.get_nodes(net_index)


                    flag_add_rxn_err = 0
                    dummy_node_id_index = 0
                    for i in range (numRxns):
                        src = []
                        dst = []
                        mod = []
                        temp_id = Rxns_ids[i]
                        try: 
                            kinetics = model.getRateLaw(i)
                        except:
                            kinetics = ""
                        rct_num = model.getNumReactants(i)
                        prd_num = model.getNumProducts(i)
                        mod_num = model.getNumModifiers(temp_id)

                        for j in range(rct_num):
                            rct_id = model.getReactant(temp_id,j)
                            for k in range(numNodes):
                                if allNodes[k].id == rct_id:
                                    src.append(allNodes[k].index)

                        for j in range(prd_num):
                            prd_id = model.getProduct(temp_id,j)
                            for k in range(numNodes):
                                if allNodes[k].id == prd_id:
                                    dst.append(allNodes[k].index)                   

                        modifiers = model.getListOfModifiers(temp_id)
                        for j in range(mod_num):
                            mod_id = modifiers[j]
                            for k in range(numNodes):
                                if allNodes[k].id == mod_id:
                                    mod.append(allNodes[k].index) 
                        mod = set(mod)

                        try: 
                            src_corr = []
                            [src_corr.append(x) for x in src if x not in src_corr]

                            dst_corr = []
                            [dst_corr.append(x) for x in dst if x not in dst_corr]

                            if len(src_corr) == 0:
                                temp_node_id = "dummy" + str(dummy_node_id_index)                   
                                comp_node_id = allNodes[dst_corr[0]].id 
                                # assume the dummy node is in the same compartment as the first src/dst node
                                comp_id = model.getCompartmentIdSpeciesIsIn(comp_node_id)
                                nodeIdx_temp = api.add_node(net_index, id=temp_node_id, size=Vec2(60,40), floating_node = True,
                                position=Vec2(40 + math.trunc (_random.random()*800), 40 + math.trunc (_random.random()*800)),
                                fill_color=api.Color(spec_fill_color[0],spec_fill_color[1],spec_fill_color[2],spec_fill_color[3]),
                                border_color=api.Color(spec_border_color[0],spec_border_color[1],spec_border_color[2],spec_border_color[3]),
                                border_width=spec_border_width, shape_index=shapeIdx)
                                
                                src_corr.append(nodeIdx_temp)
                                dummy_node_id_index += 1

                                for i in range(numComps):
                                    if comp_id == Comps_ids[i]:
                                        api.set_compartment_of_node(net_index=net_index, node_index=nodeIdx_temp, comp_index=i)
                            
                            if len(dst_corr) == 0:
                                temp_node_id = "dummy" + str(dummy_node_id_index)                   
                                comp_node_id = allNodes[src_corr[0]].id
                                comp_id = model.getCompartmentIdSpeciesIsIn(comp_node_id)
                                nodeIdx_temp = api.add_node(net_index, id=temp_node_id, size=Vec2(60,40), floating_node = True,
                                position=Vec2(40 + math.trunc (_random.random()*800), 40 + math.trunc (_random.random()*800)),
                                fill_color=api.Color(spec_fill_color[0],spec_fill_color[1],spec_fill_color[2],spec_fill_color[3]),
                                border_color=api.Color(spec_border_color[0],spec_border_color[1],spec_border_color[2],spec_border_color[3]),
                                border_width=spec_border_width, shape_index=shapeIdx)
                                
                                dst_corr.append(nodeIdx_temp)
                                dummy_node_id_index += 1

                                for i in range(numComps):
                                    if comp_id == Comps_ids[i]:
                                        api.set_compartment_of_node(net_index=net_index, node_index=nodeIdx_temp, comp_index=i)
        
                            idx = api.add_reaction(net_index, id=temp_id, reactants=src_corr, products=dst_corr,
                            fill_color=api.Color(reaction_line_color[0],reaction_line_color[1],reaction_line_color[2],reaction_line_color[3]),
                            line_thickness=reaction_line_width, modifiers = mod)
                            api.update_reaction(net_index, idx, ratelaw = kinetics,
                            fill_color=api.Color(reaction_line_color[0],reaction_line_color[1],reaction_line_color[2],reaction_line_color[3]))


                            #set the information for handle positions, center positions and use bezier as default
                        
                        except:
                            flag_add_rxn_err = 1


                        # if flag_add_rxn_err == 1:
                        #     wx.MessageBox("There are errors while loading this SBML file!", "Message", wx.OK | wx.ICON_INFORMATION)
                        
            except:
                if showDialogues:
                    wx.MessageBox("Imported SBML file is invalid.", "Message", wx.OK | wx.ICON_INFORMATION)


            






'''
Network Simulation Plugin
'''

from rkviewer_plugins import exportSBML
import wx
import wx.lib.newevent
from rkviewer.plugin.classes import PluginMetadata, WindowedPlugin, PluginCategory
from rkviewer.plugin import api
from rkviewer.plugin.api import Node, Vec2, Reaction, Color
from rkviewer.canvas.geometry import Rect
import os
from libsbml import *
import tellurium as te
import roadrunner
import simplesbml 
import time
import numpy as np
import re

from rkviewer.events import bind_handler, unbind_handler, DidPaintCanvasEvent, DidNewNetworkEvent

class SimulateModel(WindowedPlugin): 
    metadata = PluginMetadata(
        name='Simulate Model',
        author='Claire Samuels',
        version='0.0.7',
        short_desc='Simulate a reaction network.',
        long_desc='Simulate a reaction network using roadrunner and visualize the results.',
        category=PluginCategory.ANALYSIS
    )
    def create_window(self, dialog):
        ''' 
          Create a window to start reaction
          Args:
              self
              dialog
        '''
        self.dialog = dialog
        self.window = wx.ScrolledWindow(dialog, size=(370, 550))
        self.window.SetScrollbars(1,1,1,1)
        self.vsizer = wx.BoxSizer(wx.VERTICAL)

        # need to load in current model as sbml to be able to set initial concentrations/values
        # requires exportSBML version 0.3.5 or later 
        def version_err_window():
            self.window = wx.Panel(dialog, pos=(5,100), size=(300, 320))
            txt = wx.StaticText(self.window, -1, "SimulateReaction requires ExportSBML version 0.3.5 or later!", (10,10))
            txt.Wrap(250)
            return self.window

        v = exportSBML.ExportSBML.metadata.version.split(".")
        exportSBMLvers = 0
        min_version = [0,3,5]
        for i in range(3):
            if int(v[i]) < min_version[i]:
                return version_err_window()
            elif int(v[i]) > min_version[i]:
                break
       
        self.sbmldoc = exportSBML.ExportSBML.NetworkToSBML(self)

        # create tellurium model
        try:
            ant = te.sbmlToAntimony(self.sbmldoc)
            self.r = te.loada(ant)
        except BaseException as err: # TODO remove printing 
            wx.MessageBox('Invalid network model. Could not simulate.', 'Error', wx.OK | wx.ICON_INFORMATION)
            return self.window

        # create window
        self.main_sizer = wx.BoxSizer(orient=wx.VERTICAL)
        self.grid = wx.GridBagSizer(vgap=10, hgap=5)
        self.sim_grid = wx.GridBagSizer(vgap=5, hgap=5)
        
        # simulation
        '''
        self.sim_grid.Add(wx.StaticText(self.window, -1, label="Start"), pos=(1,1))
        self.sim_start_input = wx.TextCtrl(self.window, -1, value="0", size=(40,20))
        self.sim_grid.Add(self.sim_start_input, pos=(1,2))
        self.sim_grid.Add(wx.StaticText(self.window, -1, label="End"), pos=(2,1))
        self.sim_end_input = wx.TextCtrl(self.window, -1, value="20", size=(40,20))
        self.sim_grid.Add(self.sim_end_input, pos=(2,2))
        '''
        self.sim_grid.Add(wx.StaticText(self.window, -1, label="Step Size"), pos=(1,1))
        self.sim_step_size = wx.TextCtrl(self.window, -1, value="1", size=(40,20))
        self.sim_grid.Add(self.sim_step_size, pos=(1,2))

        self.sim_grid.Add(wx.StaticText(self.window, -1, label="Updates per Second"), pos=(1,3))
        self.sim_step_time = wx.TextCtrl(self.window, -1, value="1.0", size=(40,20))
        self.sim_grid.Add(self.sim_step_time, pos=(1,4))

        self.reset_btn = wx.Button(self.window, -1, "Reset", (2,1))
        self.reset_btn.Bind(wx.EVT_BUTTON, self.reset)
        self.sim_grid.Add(self.reset_btn, pos=(2,1))

        self.go_btn = wx.Button(self.window, -1, "Simulate", (2,3))
        self.go_btn.Bind(wx.EVT_BUTTON, self.go)
        self.sim_grid.Add(self.go_btn, pos=(2,3))

        self.stop_btn = wx.Button(self.window, -1, "Stop", (2,2))
        self.stop_btn.Bind(wx.EVT_BUTTON, self.stop_sim)
        self.sim_grid.Add(self.stop_btn, pos=(2,2))
        self.stop_btn.Disable()

        self.sim_time = wx.StaticText(self.window, -1, label="")
        self.sim_grid.Add(self.sim_time, pos=(3,3))

        self.main_sizer.Add(self.sim_grid)
        self.main_sizer.AddSpacer(5)

        self.param_iv_tc = []
        
        row = 1
        
        self.grid.Add(wx.StaticText(self.window, -1, label="Parameter Initial Values"), pos=(row,1), span=(1,2))
        row +=1
        param_ids = self.r.getGlobalParameterIds()
        param_values = self.r.getGlobalParameterValues() # TODO these are guarenteed to be in same order, but what if param values are not set? will there always be a value?
        for pidx, p in enumerate(param_ids):
            self.grid.Add(wx.StaticText(self.window, -1, label=p, size=(40,20)), pos=(row,1))
            pval = param_values[i]
            tc_slider = SliderWithText(self.window, init_val=pval, name=p)
            self.param_iv_tc.append(tc_slider)
            self.grid.Add(tc_slider, pos=(row, 2), span=(1,2))
            row += 1

        #self.initial_value_tc = self.species_iv_tc + self.param_iv_tc
        self.initial_value_tc = self.param_iv_tc


        # populate tc - could make a seperate function for this?

        # for painting and updating model
        self.nodeinfo = dict()
        self.max_conc = 0
        # maps ids (names) to indices and display text. points is the quantity at all the time points
        for n in api.get_nodes(0):
            self.nodeinfo[n.id] = {"index": n.index, "points": [], "floating": n.floating_node}
            if n.floating_node:
                self.nodeinfo[n.id]["points"] = [n.concentration]
                if n.concentration > self.max_conc:
                    self.max_conc = n.concentration
            # TODO do first point here by pulling th concentration from the node
        self.reacinfo = dict()
        for r in api.get_reactions(0):
            self.reacinfo[r.id] = {"index": r.index, "rate_law": ""}

        self.main_sizer.Add(self.grid)

        self.window.SetSizer(self.main_sizer)

        self.paint_id = bind_handler(DidPaintCanvasEvent, self.on_paint)
        bind_handler(DidNewNetworkEvent, self.remove)
        self.dialog.Bind(wx.EVT_CLOSE, self.on_close)
        
        self.time = 0
        self.timer = wx.Timer(owner=self.dialog)
        return self.window

    def stop_sim(self, evt):
        '''
        Stops the simulation
        '''
        #self.chose_stop_sim = True
        self.stop_btn.Disable()
        self.go_btn.Enable()
        try:
            self.timer.Stop()
        except: # timer hasn't been initialized
            pass
        #self.remove(any) # TODO change this. remove should only happen when the window is closed

        # going to update the network every time the simulation is stopped or "reset" is chosen
        self.update_network(0)
        #donemsg = wx.messagebox("done.", "simulate network", wx.ok | wx.cancel)
        #self.update_network(0, donemsg==wx.ok)
        #if donemsg == wx.ok:
        #    self.update_network(0, true)

    def on_close(self, evt):
        self.stop_sim(evt)
        self.remove(evt)
        self.dialog.Destroy()

    def reset(self, evt):
        '''
        handler for "reset" button"
        '''
        self.r.reset()
        self.time = 0.0
        self.sim_time.SetLabel('Time: {}'.format(str(round(self.time,3)))) #TODO
        self.update_node_info()


    def go(self, evt):

        # handler for go button

        # validate input
        param_indices = []
        param_values = []
        for idx, sp in enumerate(self.initial_value_tc):
            name = sp.GetName()
            if sp.IsModified():
                # check that input is valid
                inpt = sp.GetValue()
                try:
                    f_inpt = float(inpt) # change to float, then change back...
                except ValueError:
                    wx.MessageBox("Invalid input for {}".format(name), "message", wx.OK | wx.ICON_INFORMATION)
                    return
                #self.model.addinitialassignment(name, str(f_inpt))
                #self.r.setValue('init({})'.format(name), f_inpt)
                param_indices.append(idx)
                param_values.append(f_inpt)
        if len(param_values)>0:
            self.r.model.setGlobalParameterValues(param_indices, param_values)

                
        '''
        # update rate laws
        for r in self.model.getlistofreactionids():
            self.reacinfo[r]["rate_law"] = self.model.getratelaw(r) 
        ''' # todo do not have to use rate laws at all

        self.go_btn.Disable()
        self.stop_btn.Enable()
        #self.chose_stop_sim = false
        
        self.step_time = 1.0 / float(self.sim_step_time.GetValue()) # ex: step time of 0.2 <=> 5 steps will occur per second
        self.step_size = float(self.sim_step_size.GetValue())
        self.time = self.r.oneStep(self.time, self.step_size) # note that self.time and self.timer are completely seperate
        self.sim_time.SetLabel('Time: {}'.format(str(round(self.time, 3))))
        self.update_node_info()
        self.dialog.Bind(wx.EVT_TIMER, self.do_one_step, self.timer)

        self.timer.Start(1000 * self.step_time)

    def do_one_step(self, evt):
        '''
        handler for timer event
        '''
        self.time = self.r.oneStep(self.time, self.step_size)
        self.sim_time.SetLabel('Time: {}'.format(str(round(self.time,3))))
        self.update_node_info()
        # api.update_canvas() # todo occurs when window is closed

    def update_node_info(self): 
        ''' 
        update concentrations to be displayed by on_paint
        '''
        net_index = 0
        sim = self.r.getFloatingSpeciesConcentrationsNamedArray()
        # want the final value of the concentration for the nodes
        # get the number of columns
        for i in range(len(sim.colnames)):
            col = sim[:,i]
            node_id = sim.colnames[i]
            if node_id in self.nodeinfo: # todo this should always be true
                self.nodeinfo[node_id]["points"].extend(col)

    def update_network(self, net_index):
        '''
        update model node concentrations and reaction rate laws. 
        '''
        with api.group_action():
            for node_id in self.nodeinfo:
                idx = self.nodeinfo[node_id]["index"]
                #if len(self.nodeinfo[node_id]["points"]) > 0:
                if self.nodeinfo[node_id]["floating"]:
                    final_conc = self.nodeinfo[node_id]["points"][-1]
                    api.update_node(net_index, idx, concentration=final_conc)
            '''
            for reac_id in self.reacinfo:
                idx = self.reacinfo[reac_id]["index"]
                api.update_reaction(net_index, idx, ratelaw=self.reacinfo[reac_id]["rate_law"])
            '''

    def remove(self, evt):
        try:
            unbind_handler(self.paint_id) # todo this should happen in events.py, not here
        except KeyError:
            pass

    def on_paint(self, evt):


        # paint node concentrations next to the nodes 
        gc = evt.gc
        # todo have an option between numbers and shapes
        '''
        font = wx.font(wx.fontinfo(10))
        gc.setfont(font, wx.colour(0,0,0))
        pen1 = gc.createpen(wx.graphicspeninfo(wx.colour(0,0,150)))
        gc.setpen(pen1)

        net_index = 0
        for node_id in self.nodeinfo:
            info = self.nodeinfo[node_id]
            node = api.get_node_by_index(net_index, info["index"])
            gc.drawtext(str(round(info["points"][self.current_point], 3)), node.position.x - 30, node.position.y + 50)
        '''
        # create pens
        pen1 = gc.CreatePen(wx.GraphicsPenInfo(wx.Colour(170,170,170)))
        brush1 = wx.Brush(wx.Colour(170,170,170), style = wx.BRUSHSTYLE_SOLID)
        pen2 = gc.CreatePen(wx.GraphicsPenInfo(wx.Colour(0, 120, 215)))
        brush2 = wx.Brush(wx.Colour(30, 120, 215), style = wx.BRUSHSTYLE_SOLID)
        yellowpen = gc.CreatePen(wx.GraphicsPenInfo(wx.Colour(250,230,0)))
        yellowbrush = wx.Brush(wx.Colour(250,230,0), style = wx.BRUSHSTYLE_SOLID)
        boundarypen = gc.CreatePen(wx.GraphicsPenInfo(wx.Colour(250,200,3)))
        boundarybrush = wx.Brush(wx.Colour(250,200,3), style = wx.BRUSHSTYLE_SOLID)
        # blue brush
        bg_w = 20
        bg_h = 50
        bg_x = 30 
        bg_y = 0
        max_conc = 20 # this is pretty arbitraty, could change
        if self.max_conc > 0:
            max_conc = self.max_conc * 3

        def rect_from_conc(concentration):
            '''# calculate dimensions of bar for graph
            '''
            # 4 px buffer
            w = bg_w - 8
            h = bg_h - 8
            x = bg_x - 4
            y = bg_y - 4

            ratio =  (concentration)/max_conc
            adj_h = round(h * ratio)
            adj_y = y + adj_h - h # maybe switch adj_h and h

            return Rect(Vec2(x,adj_y), Vec2(w,adj_h))
    
        net_index = 0
        for node_id in self.nodeinfo:
            info = self.nodeinfo[node_id]
            #try:
            node = api.get_node_by_index(net_index, info["index"])
            #if len(info["points"]) > 0:
            if info["floating"]:
                # draw background
                gc.SetPen(pen1)
                gc.SetBrush(brush1)
                gc.DrawRoundedRectangle(node.position.x - bg_x, node.position.y - bg_y - (bg_h - node.size.y), bg_w, bg_h, 2)
                conc = info["points"][-1]
                # draw dynamic bar
                if conc <= max_conc:
                    gc.SetBrush(brush2)
                    gc.SetPen(pen2)
                    bar = rect_from_conc(conc)
                else:
                    gc.SetBrush(yellowbrush)
                    gc.SetPen(yellowpen)
                    bar = rect_from_conc(max_conc)
                gc.DrawRoundedRectangle(node.position.x - bar.position.x, node.position.y - bar.position.y - (bg_h - node.size.y), bar.size.x, bar.size.y, 0)
            #except:
            #    pass # TODO handle
            '''
            else: # boundary node
                # draw background
                gc.SetPen(pen1)
                gc.SetBrush(brush1)
                gc.DrawRoundedRectangle(node.position.x - bg_x, node.position.y - bg_y, bg_w, bg_h, 2)
                conc = info["points"][-1]
                # draw dynamic bar
                if conc <= max_conc:
                    gc.SetBrush(brush2)
                    gc.SetPen(pen2)
                    bar = rect_from_conc(conc)
                else:
                    gc.SetBrush(yellowbrush)
                    gc.SetPen(yellowpen)
                    bar = rect_from_conc(max_conc)
                gc.DrawRoundedRectangle(node.position.x - bar.position.x, node.position.y - bar.position.y, bar.size.x, bar.size.y, 0)
            '''


class SliderWithText(wx.Panel):
    def __init__(self, parent, init_val=0.0, name=None):
        wx.Panel.__init__(self, parent, size=(200,40))
        self.name=name
        self.slider = wx.Slider(self, -1, size=(150,20), maxValue=100)
        self.tc = wx.TextCtrl(self, -1, size=(30,20))
        self.sizer = wx.GridBagSizer()
        self.SetValue(init_val)

        self.sizer.Add(self.tc, pos=(0,0), border=5)
        self.sizer.Add(self.slider, pos=(0,1), border=20, flag=wx.RIGHT)
        self.labels_sizer = wx.BoxSizer()
        self.labels_sizer.Add(wx.StaticText(self, -1, "0", size=(10,15)), border=10, flag=wx.LEFT)
        self.labels_sizer.AddSpacer(115)
        self.labels_sizer.Add(wx.StaticText(self, -1, "20", size=(15,15)))
        self.sizer.Add(self.labels_sizer, pos=(1,1))
        self.SetSizer(self.sizer)
        
        self._is_modified = False
        self.slider.Bind(wx.EVT_SLIDER, self._on_slider)
        self.tc.Bind(wx.EVT_TEXT, self._on_text)

    def SetValue(self, value):
        self.tc.SetValue(str(value))
        self._set_slider_val(value)
        self._is_modified = True

    def GetValue(self):
        return float(self.tc.GetValue())

    def GetName(self):
        return self.name

    def IsModified(self):
        return self._is_modified

    def _on_slider(self, evt):
        self.tc.SetValue(str(self._get_slider_val()))
        self._is_modified=True
    def _on_text(self, evt):

        try:
            val = float(self.tc.GetValue())
        except ValueError as err: # TODO implement
            return
        self._set_slider_val(val)
        self._is_modified=True
        
    def _get_slider_val(self):
        return self.slider.GetValue()/5

    def _set_slider_val(self, value):
        self.slider.Enable()
        if value <= 20 and value >= 0:
            self.slider.SetValue(round(value*5))
        else:
            self.slider.Disable()



'''
Import a directory of SBML and Antimony files, visualize reactions, capture and save images.

Version 0.0.4: Author: Claire Samuels (2021)
'''

try:
    from rkviewer_plugins import importSBML
except:
    import importSBML

import wx
from rkviewer.plugin.classes import PluginMetadata, WindowedPlugin, PluginCategory
from rkviewer.plugin import api
from rkviewer.plugin.api import Node, Vec2, Reaction, Color
import os
from libsbml import *
import tellurium

class CaptureSBML(WindowedPlugin):
      metadata = PluginMetadata(
        name='CaptureSBML',
        author='Claire Samuels',
        version='0.0.4',
        short_desc='Visualize and capture SBML or Antimony.',
        long_desc='Import a directory of SBML and Antimony files, visualize reactions, capture and save images.',
        category=PluginCategory.ANALYSIS
    )
      def create_window(self, dialog):
        """
        Create a window to export the SBML.
        Args:
            self
            dialog
        """
        # requires importSBML version 0.0.3
        def version_err_window():
            self.window = wx.Panel(dialog, pos=(5,100), size=(300, 320))
            txt = wx.StaticText(self.window, -1, "CaptureSBML requires ImportSBML version 0.0.3 or later!", (10,10))
            txt.Wrap(250)
            return self.window

        v = importSBML.IMPORTSBML.metadata.version.split(".")
        min_version = [0,0,3]
        for i in range(3):
            if int(v[i]) < min_version[i]:
                return version_err_window()
            elif int(v[i]) > min_version[i]:
                break

        # import button
        self.window = wx.Panel(dialog, pos=(5,100), size=(300, 320))
        import_btn = wx.Button(self.window, -1, 'Import', (5, 5))
        import_btn.Bind(wx.EVT_BUTTON, self.Import)

        # Directory text:
        self.display_dir = wx.StaticText(self.window, -1, '', (85,10), (180,25),
                                         style= wx.ST_NO_AUTORESIZE | wx.ST_ELLIPSIZE_START)

        # files to import
        wx.StaticText(self.window, -1, 'Files to Display:', (5,30))
        self.selectedFiles = wx.TextCtrl(self.window, -1, "", (10,50), (240, 150),
                                         style=wx.TE_MULTILINE | wx.TE_READONLY)

        # directory to save images to
        output_dir_label = wx.StaticText(self.window, -1, "Output Directory:", (5,215))
        self.output_dir = wx.TextCtrl(self.window, -1, "", (100,210), (180,25))

        # go button
        go_btn = wx.Button(self.window, -1, 'Go', (200,250))
        go_btn.Bind(wx.EVT_BUTTON, self.Go)

        return self.window

      def Import(self, evt):
        '''
        Handler for "Import" button
        Choose a directory to iterate through'''
        self.dirname = ""
        dlg = wx.DirDialog(self.window, "Choose a directory", self.dirname,
                            wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST)
        self.file_list = []
        if dlg.ShowModal() == wx.ID_OK:
            self.dirname = dlg.GetPath()
            for file in os.listdir(self.dirname):
              # only want .xml files and .ant
              filename = os.fsdecode(file)
              if filename.endswith(".xml") | filename.endswith(".ant"):
                self.file_list.append(filename)
        self.display_dir.SetLabel(str(self.dirname))
        self.selectedFiles.SetLabel(str(self.file_list))

        self.output_dir.Clear()
        self.output_dir.write('{}\\reaction_visualizations'.format(self.dirname))
        dlg.Destroy()

      def Go(self, evt):
        '''
        Handler for "Go" button.
        Reads each file in selected directory. If readable SBML, displays on canvas'''
        reader = SBMLReader()

        self.output_dir_path = self.output_dir.Value
        if not os.path.isdir(self.output_dir_path):
          os.mkdir(self.output_dir_path)
        else:
          dir_cnt = 1
          while os.path.isdir('{} ({})'.format(self.output_dir_path, dir_cnt)):
            dir_cnt += 1
          self.output_dir_path = '{} ({})'.format(self.output_dir_path, dir_cnt)
          os.mkdir(self.output_dir_path)

        self.blank_canvas = []

        for filename in self.file_list:

          f = open(os.path.join(self.dirname, filename), 'r')
          sbmlStr = f.read()

          # convert .ant files to .xml
          if filename.endswith(".ant"):
            try:
              sbmlStr = tellurium.tellurium.antimonyToSBML(sbmlStr)
            except:
              pass

          doc = reader.readSBMLFromString(sbmlStr)
          if doc.getNumErrors() > 0:
            self.blank_canvas.append(filename)
          else:
            try:
              importSBML.IMPORTSBML.DisplayModel(self, sbmlStr, False, True)
            except ValueError:
              pass
            self.Capture(filename)
        if len(self.blank_canvas) > 0:
          flist = ""
          for blkf in self.blank_canvas:
            flist += blkf + ", "
          flist = flist[:-2]
          wx.MessageBox("Done. Images saved to\n{}\nWarning: No node or reaction information found for {}.\nNo visualizations saved for these files.".format(self.output_dir_path, flist),
                         "Message", wx.OK | wx.ICON_INFORMATION)
        else:
          wx.MessageBox("Done. Images saved to\n{}".format(self.output_dir_path), "Message",
                         wx.OK | wx.ICON_INFORMATION)

      def Capture(self, filename):
        '''
        Saves a png of the current canvas to the chosen output directory
        Args:
          self
          filename: name of sbml file currently displayed'''

        # ensure that the canvas isn't blank
        if api.node_count(0) == 0 and api.reaction_count(0) == 0 and api.compartments_count(0) <= 1:
          self.blank_canvas.append(filename)
        else:
          pathname = os.path.join(self.output_dir_path, 'img_{}.png'.format(filename))

          canv = api.get_canvas()
          img = canv.DrawActiveRectToImage()

          img.SaveFile(pathname)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rkviewer',
 'rkviewer.canvas',
 'rkviewer.plugin',
 'rkviewer.resources',
 'rkviewer_plugins']

package_data = \
{'': ['*']}

install_requires = \
['commentjson>=0.9.0,<0.10.0',
 'marshmallow-polyfield>=5.10,<6.0',
 'marshmallow>=3.11.1,<4.0.0',
 'sortedcontainers>=2.3.0,<3.0.0',
 'wxPython==4.1.1']

extras_require = \
{':python_version >= "3.6" and python_version < "3.7"': ['dataclasses>=0.8,<0.9'],
 'sbml': ['tellurium>=2.2.1,<3.0.0',
          'sbml2matlab==1.2.3',
          'networkx>=2.5.1,<3.0.0',
          'simplesbml>=2.2.0,<3.0.0',
          'python-libsbml>=5.18.0,<6.0.0',
          'pandas>=1.3.0,<2.0.0']}

entry_points = \
{'console_scripts': ['coyote = rkviewer.main:main']}

setup_kwargs = {
    'name': 'coyote-gui',
    'version': '0.1b8',
    'description': 'Extensible visualization and editing tool for reaction networks.',
    'long_description': None,
    'author': 'Gary Geng',
    'author_email': 'xiaog3@uw.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)
